@@ -32,6 +32,7 @@
 
 #include <libtpms/tpm_nvfilename.h>
 
+#include "profile.h"
 #include "swtpm.h"
 #include "swtpm_conf.h"
 #include "swtpm_utils.h"
@@ -60,6 +61,7 @@
 #define SETUP_DECRYPTION_F          (1 << 14)
 #define SETUP_WRITE_EK_CERT_FILES_F (1 << 15)
 #define SETUP_RECONFIGURE_F         (1 << 16)
+#define SETUP_RSA_KEYSIZE_BY_USER_F (1 << 17)
 
 /* default configuration file */
 #define SWTPM_SETUP_CONF "swtpm_setup.conf"
@@ -147,7 +149,7 @@ static int tpm_get_specs_and_attributes(struct swtpm *swtpm, gchar ***params)
             else
                 str = g_strdup(json_reader_get_string_value(jr));
 
-            *params = concat_arrays(*params,
+            *params = concat_varrays(*params,
                                     (gchar*[]){
                                         g_strdup(parser_rules[idx].optname),
                                         str,
@@ -183,7 +185,7 @@ static int call_create_certs(unsigned long flags, unsigned int cert_flags,
     g_autofree gchar *create_certs_tool = NULL;
     g_autofree gchar *create_certs_tool_config = NULL;
     g_autofree gchar *create_certs_tool_options = NULL;
-    g_autofree gchar **cmd = NULL;
+    g_autofree const gchar **cmd = NULL;
     gchar **params = NULL; /* must free */
     g_autofree gchar *prgname = NULL;
     gboolean success;
@@ -215,28 +217,28 @@ static int call_create_certs(unsigned long flags, unsigned int cert_flags,
         }
 
         if (flags & SETUP_TPM2_F) {
-            params = concat_arrays(params,
+            params = concat_varrays(params,
                                 (gchar*[]){
                                     g_strdup("--tpm2"),
                                     NULL
                                 }, TRUE);
         }
-        cmd = concat_arrays((gchar*[]) {
+        cmd = concat_arrays((const gchar*[]) {
                                 create_certs_tool_path,
                                 "--type", "_",  /* '_' must be at index '2' ! */
-                                "--ek", (gchar *)ekparam,
-                                "--dir", (gchar *)certsdir,
+                                "--ek", ekparam,
+                                "--dir", certsdir,
                                 NULL
                             }, NULL, FALSE);
         if (gl_LOGFILE != NULL)
-            cmd = concat_arrays(cmd, (gchar*[]){"--logfile", (gchar *)gl_LOGFILE, NULL}, TRUE);
+            cmd = concat_arrays(cmd, (const gchar*[]){"--logfile", gl_LOGFILE, NULL}, TRUE);
         if (vmid != NULL)
-            cmd = concat_arrays(cmd, (gchar*[]){"--vmid", (gchar *)vmid, NULL}, TRUE);
-        cmd = concat_arrays(cmd, params, TRUE);
+            cmd = concat_arrays(cmd, (const gchar*[]){"--vmid", vmid, NULL}, TRUE);
+        cmd = concat_arrays(cmd, (const char **)params, TRUE);
         if (create_certs_tool_config != NULL)
-            cmd = concat_arrays(cmd, (gchar*[]){"--configfile", create_certs_tool_config, NULL}, TRUE);
+            cmd = concat_arrays(cmd, (const gchar*[]){"--configfile", create_certs_tool_config, NULL}, TRUE);
         if (create_certs_tool_options != NULL)
-            cmd = concat_arrays(cmd, (gchar*[]){"--optsfile", create_certs_tool_options, NULL}, TRUE);
+            cmd = concat_arrays(cmd, (const gchar*[]){"--optsfile", create_certs_tool_options, NULL}, TRUE);
 
         s = g_strrstr(create_certs_tool, G_DIR_SEPARATOR_S);
         if (s)
@@ -253,12 +255,12 @@ static int call_create_certs(unsigned long flags, unsigned int cert_flags,
 
                 cmd[2] = (gchar *)flags_to_certfiles[idx].type; /* replaces the "_" above */
 
-                s = g_strjoinv(" ", cmd);
+                s = g_strjoinv(" ", (char **)cmd);
                 logit(gl_LOGFILE, "  Invoking %s\n", s);
                 g_free(s);
 
-                success = g_spawn_sync(NULL, cmd, NULL, 0, NULL, NULL,
-                                       &standard_output, &standard_error, &exit_status, &error);
+                success = spawn_sync(NULL, cmd, NULL, 0, NULL, NULL,
+                                     &standard_output, &standard_error, &exit_status, &error);
                 if (!success) {
                     logerr(gl_LOGFILE, "An error occurred running %s: %s\n",
                            create_certs_tool, error->message);
@@ -406,8 +408,8 @@ static int tpm2_create_ek_and_cert(unsigned long flags, const gchar *config_file
                                    unsigned int rsa_keysize, struct swtpm2 *swtpm2,
                                    const gchar *user_certsdir)
 {
+    const char *key_description = "";
     g_autofree gchar *ekparam = NULL;
-    const char *key_description;
     unsigned long cert_flags;
     const gchar *key_type;
     size_t idx;
@@ -471,15 +473,9 @@ static int tpm2_create_eks_and_certs(unsigned long flags, const gchar *config_fi
 /* Get the default PCR banks from the config file and if nothing can
    be found there use the DEFAULT_PCR_BANKS #define.
  */
-static gchar *get_default_pcr_banks(const gchar *config_file)
+static gchar *get_default_pcr_banks(gchar *const *config_file_lines)
 {
-    g_auto(GStrv) config_file_lines = NULL;
     gchar *pcr_banks;
-    int ret;
-
-    ret = read_file_lines(config_file, &config_file_lines);
-    if (ret != 0)
-        return NULL;
 
     pcr_banks = get_config_value(config_file_lines, "active_pcr_banks");
     if (pcr_banks)
@@ -491,6 +487,51 @@ static gchar *get_default_pcr_banks(const gchar *config_file)
     return pcr_banks;
 }
 
+/* Get the default RSA keysize from the config file */
+static gchar *get_default_rsa_keysize(gchar *const *config_file_lines)
+{
+    gchar *rsa_keysize;
+
+    rsa_keysize = get_config_value(config_file_lines, "rsa_keysize");
+    if (rsa_keysize)
+        g_strstrip(rsa_keysize);
+    if (rsa_keysize == NULL || strlen(rsa_keysize) == 0) {
+        g_free(rsa_keysize);
+        rsa_keysize = g_strdup_printf("%d", DEFAULT_RSA_KEYSIZE);
+    }
+    return rsa_keysize;
+}
+
+/* Get the default profile from the config file */
+static gchar *get_default_profile(gchar *const *config_file_lines)
+{
+    gchar *profile;
+
+    profile = get_config_value(config_file_lines, "profile");
+    if (profile)
+        g_strstrip(profile);
+    return profile;
+}
+
+/* If available, open the default profile and return its file descriptor */
+static int get_default_profile_fd(gchar *const *config_file_lines)
+{
+    g_autofree gchar *profile_file = NULL;
+    int fd;
+
+    profile_file = get_config_value(config_file_lines, "profile_file");
+    if (!profile_file)
+        return -1;
+
+    fd = open(profile_file, O_RDONLY);
+    if (fd < 0) {
+        logerr(gl_LOGFILE, "Could not read default profile '%s': %s",
+               profile_file, strerror(errno));
+        return -2;
+    }
+    return fd;
+}
+
 /* Activate the given list of PCR banks. If pcr_banks is '-' then leave
  * the configuration as-is.
  */
@@ -526,19 +567,49 @@ static int tpm2_activate_pcr_banks(struct swtpm2 *swtpm2,
     return 0;
 }
 
+static int log_active_profile(struct swtpm2 *swtpm2)
+{
+    g_autofree gchar *profile = NULL;
+    char *tmp;
+
+    profile = swtpm2->ops->get_active_profile(&swtpm2->swtpm);
+    if (!profile) {
+        logerr(gl_LOGFILE, "Could not get active profile.\n");
+        return 1;
+    }
+    /* Strip out surrounding '{"ActiveProfile":<to display>} */
+    tmp = strrchr(profile, '}');
+    if (!tmp)
+        goto malformatted;
+    *tmp = 0;
+
+    tmp = strchr(profile, ':');
+    if (!tmp)
+        goto malformatted;
+
+    logit(gl_LOGFILE, "Active profile: %s\n", tmp + 1);
+    return 0;
+
+malformatted:
+    logerr(gl_LOGFILE, "Malformatted active profile");
+    return 1;
+}
+
 /* Simulate manufacturing a TPM 2: create keys and certificates */
 static int init_tpm2(unsigned long flags, gchar **swtpm_prg_l, const gchar *config_file,
                      const gchar *tpm2_state_path, const gchar *vmid, const gchar *pcr_banks,
                      const gchar *swtpm_keyopt, int *fds_to_pass, size_t n_fds_to_pass,
                      unsigned int rsa_keysize, const gchar *certsdir,
-                     const gchar *user_certsdir)
+                     const gchar *user_certsdir, const gchar *json_profile,
+                     int json_profile_fd, const gchar *profile_remove_disabled_param)
 {
     struct swtpm2 *swtpm2;
     struct swtpm *swtpm;
     int ret;
 
     swtpm2 = swtpm2_new(swtpm_prg_l, tpm2_state_path, swtpm_keyopt, gl_LOGFILE,
-                        fds_to_pass, n_fds_to_pass);
+                        fds_to_pass, n_fds_to_pass, json_profile, json_profile_fd,
+                        profile_remove_disabled_param);
     if (swtpm2 == NULL)
         return 1;
     swtpm = &swtpm2->swtpm;
@@ -550,6 +621,10 @@ static int init_tpm2(unsigned long flags, gchar **swtpm_prg_l, const gchar *conf
     }
 
     if (!(flags & SETUP_RECONFIGURE_F)) {
+        ret = log_active_profile(swtpm2);
+        if (ret)
+            goto error;
+
         if ((flags & SETUP_CREATE_SPK_F)) {
             ret = swtpm2->ops->create_spk(swtpm, !!(flags & SETUP_TPM2_ECC_F), rsa_keysize);
             if (ret != 0)
@@ -769,19 +844,19 @@ error:
  * This function returns 2 if the state exists but flag is set to not to overwrite it,
  * 0 in case we can overwrite it, 1 if the state exists.
  */
-static int check_state_overwrite(gchar **swtpm_prg_l, unsigned int flags,
+static int check_state_overwrite(const gchar **swtpm_prg_l, unsigned int flags,
                                  const char *tpm_state_path)
 {
     gboolean success;
     g_autofree gchar *standard_output = NULL;
     int exit_status = 0;
     g_autoptr(GError) error = NULL;
-    g_autofree gchar **argv = NULL;
+    g_autofree const gchar **argv = NULL;
     g_autofree gchar *statearg = g_strdup_printf("backend-uri=%s", tpm_state_path);
     g_autofree gchar *logop = NULL;
-    g_autofree gchar **my_argv = NULL;
+    g_autofree const gchar **my_argv = NULL;
 
-    my_argv = concat_arrays((gchar*[]) {
+    my_argv = concat_arrays((const gchar*[]) {
                                 "--print-states",
                                 "--tpmstate",
                                 statearg,
@@ -789,17 +864,17 @@ static int check_state_overwrite(gchar **swtpm_prg_l, unsigned int flags,
                             }, NULL, FALSE);
 
     if (flags & SETUP_TPM2_F)
-        my_argv = concat_arrays(my_argv, (gchar*[]) { "--tpm2", NULL }, TRUE);
+        my_argv = concat_arrays(my_argv, (const gchar*[]) { "--tpm2", NULL }, TRUE);
 
     if (gl_LOGFILE != NULL) {
         logop = g_strdup_printf("file=%s", gl_LOGFILE);
-        my_argv = concat_arrays(my_argv, (gchar*[]){"--log", logop, NULL}, TRUE);
+        my_argv = concat_arrays(my_argv, (const gchar*[]){"--log", logop, NULL}, TRUE);
     }
 
     argv = concat_arrays(swtpm_prg_l, my_argv, FALSE);
 
-    success = g_spawn_sync(NULL, argv, NULL, G_SPAWN_STDERR_TO_DEV_NULL, NULL, NULL,
-                           &standard_output, NULL, &exit_status, &error);
+    success = spawn_sync(NULL, argv, NULL, G_SPAWN_STDERR_TO_DEV_NULL, NULL, NULL,
+                         &standard_output, NULL, &exit_status, &error);
     if (!success) {
         logerr(gl_LOGFILE, "Could not start swtpm '%s': %s\n", swtpm_prg_l[0], error->message);
         return 1;
@@ -886,7 +961,7 @@ static void usage(const char *prgname, const char *default_config_file)
         "--create-platform-cert\n"
         "                 : Create a platform certificate; this implies --create-ek-cert\n"
         "\n"
-        "--create-spk     : Create storage primary key; this requires --tpm2\n"
+        "--create-spk     : Create storage primary key; this requires --tpm2; deprecated\n"
         "\n"
         "--lock-nvram     : Lock NVRAM access\n"
         "\n"
@@ -961,6 +1036,28 @@ static void usage(const char *prgname, const char *default_config_file)
         "                   The active PCR banks can be changed but no new keys will\n"
         "                   be created.\n"
         "\n"
+        "--profile <json-profile>\n"
+        "                 : Configure swtpm with the given profile.\n"
+        "\n"
+        "--profile-name <profile name | built-in profile name>\n"
+        "                 : Search for a profile with the <name>.json in distro and\n"
+        "                   local directories; if not found try it as a built-in.\n"
+        "\n"
+        "--profile-file <file>\n"
+        "                 : Configure swtpm with a profile read from the given file.\n"
+        "\n"
+        "--profile-file-fd <fd>\n"
+        "                 : Configure swtpm with a profile read from a file descriptor.\n"
+        "\n"
+        "--profile-remove-disabled check|fips-host\n"
+        "                 : Instruct swtpm to remove algorithms that may be disabled by\n"
+        "                   FIPS mode on the host from 'custom' profile.\n"
+        "                   check: algorithms are tested.\n"
+        "                   fips-host: no testing.\n"
+        "\n"
+        "--print-profiles : Display all local and distro-provided profile as well as\n"
+        "                   the ones built into libtpms and exit.\n"
+        "\n"
         "--version        : Display version and exit\n"
         "\n"
         "--help,-h        : Display this help screen\n\n",
@@ -973,13 +1070,14 @@ static void usage(const char *prgname, const char *default_config_file)
         );
 }
 
-static int get_swtpm_capabilities(gchar **swtpm_prg_l, gboolean is_tpm2,
+static int get_swtpm_capabilities(const gchar **swtpm_prg_l, gboolean is_tpm2,
                                   gchar **standard_output)
 {
-    gchar *my_argv[] = { "--print-capabilities", is_tpm2 ? "--tpm2" : NULL, NULL };
+    const gchar *my_argv[] = { "--print-capabilities", is_tpm2 ? "--tpm2" : NULL, NULL };
+    g_autofree gchar *standard_error = NULL;
     g_autofree gchar *logop = NULL;
     g_autoptr(GError) error = NULL;
-    g_autofree gchar **argv = NULL;
+    g_autofree const gchar **argv = NULL;
     int exit_status = 0;
     gboolean success;
     int ret = 1;
@@ -988,22 +1086,27 @@ static int get_swtpm_capabilities(gchar **swtpm_prg_l, gboolean is_tpm2,
 
     if (gl_LOGFILE != NULL) {
         logop = g_strdup_printf("file=%s", gl_LOGFILE);
-        argv = concat_arrays(argv, (gchar*[]){"--log", logop, NULL}, TRUE);
+        argv = concat_arrays(argv, (const gchar*[]){"--log", logop, NULL}, TRUE);
     }
 
-    success = g_spawn_sync(NULL, argv, NULL, G_SPAWN_STDERR_TO_DEV_NULL, NULL, NULL,
-                           standard_output, NULL, &exit_status, &error);
+    success = spawn_sync(NULL, argv, NULL, 0, NULL, NULL,
+                         standard_output, &standard_error, &exit_status, &error);
     if (!success) {
         logerr(gl_LOGFILE, "Could not start swtpm '%s': %s\n", swtpm_prg_l[0], error->message);
         goto error;
     }
+    if (exit_status != 0) {
+        /* possible: failure to access log file */
+        logerr(gl_LOGFILE, "Failed to run swtpm '%s': %s\n", swtpm_prg_l[0], standard_error);
+        goto error;
+    }
     ret = 0;
 
 error:
     return ret;
 }
 
-static int get_supported_tpm_versions(gchar **swtpm_prg_l, gboolean *swtpm_has_tpm12,
+static int get_supported_tpm_versions(const gchar **swtpm_prg_l, gboolean *swtpm_has_tpm12,
                                       gboolean *swtpm_has_tpm2)
 {
     g_autofree gchar *standard_output = NULL;
@@ -1024,7 +1127,7 @@ static int get_supported_tpm_versions(gchar **swtpm_prg_l, gboolean *swtpm_has_t
  *  - [ 1024, 2048, 3072 ]
  *  - [] (empty array, indicating only 2048 bit RSA keys are supported)
  */
-static int get_rsa_keysizes(unsigned long flags, gchar **swtpm_prg_l,
+static int get_rsa_keysizes(unsigned long flags, const gchar **swtpm_prg_l,
                             unsigned int **keysizes, size_t *n_keysizes)
 {
     g_autofree gchar *standard_output = NULL;
@@ -1060,7 +1163,7 @@ error:
 }
 
 /* Return the RSA key size capabilities in a NULL-terminated array */
-static int get_rsa_keysize_caps(unsigned long flags, gchar **swtpm_prg_l,
+static int get_rsa_keysize_caps(unsigned long flags, const gchar **swtpm_prg_l,
                                 gchar ***keysize_strs)
 {
     unsigned int *keysizes = NULL;
@@ -1081,11 +1184,26 @@ static int get_rsa_keysize_caps(unsigned long flags, gchar **swtpm_prg_l,
     return 0;
 }
 
+static int validate_json_profile(const gchar **swtpm_prg_l, const char *json_profile)
+{
+    g_autofree gchar *standard_output = NULL;
+    int ret;
+
+    ret = get_swtpm_capabilities(swtpm_prg_l, TRUE, &standard_output);
+    if (ret)
+        return ret;
+
+    return check_json_profile(standard_output, json_profile);
+}
+
 /* Print the JSON object of swtpm_setup's capabilities */
-static int print_capabilities(char **swtpm_prg_l, gboolean swtpm_has_tpm12,
+static int print_capabilities(const char **swtpm_prg_l, gboolean swtpm_has_tpm12,
                               gboolean swtpm_has_tpm2)
 {
+    g_autofree gchar *standard_output = NULL;
     g_autofree gchar *param = g_strdup("");
+    g_autofree gchar *profile_list = NULL;
+    gchar **profile_names = NULL;
     gchar **keysize_strs = NULL;
     gchar *tmp;
     size_t i;
@@ -1101,20 +1219,43 @@ static int print_capabilities(char **swtpm_prg_l, gboolean swtpm_has_tpm12,
         param = tmp;
     }
 
+    if (swtpm_has_tpm2) {
+        ret = get_swtpm_capabilities(swtpm_prg_l, TRUE, &standard_output);
+        if (ret)
+            goto error;
+        ret = get_profile_names(standard_output, &profile_names);
+        if (ret)
+            goto error;
+
+        if (g_strv_length(profile_names) > 0) {
+            tmp = g_strjoinv("\", \"", profile_names);
+            profile_list = g_strdup_printf(" \"%s\" ", tmp);
+            g_free(tmp);
+        }
+    }
+
     printf("{ \"type\": \"swtpm_setup\", "
            "\"features\": [ %s%s\"cmdarg-keyfile-fd\", \"cmdarg-pwdfile-fd\", \"tpm12-not-need-root\""
            ", \"cmdarg-write-ek-cert-files\", \"cmdarg-create-config-files\""
            ", \"cmdarg-reconfigure-pcr-banks\""
-           "%s ], "
+           "%s"
+           ", \"cmdarg-profile\", \"cmdarg-profile-remove-disabled\""
+           ""
+           " ], "
+           "\"profiles\": [%s], "
            "\"version\": \"" VERSION "\" "
            "}\n",
            swtpm_has_tpm12 ? "\"tpm-1.2\", " : "",
            swtpm_has_tpm2  ? "\"tpm-2.0\", " : "",
-           param);
+           param,
+           profile_list ? profile_list : ""
+           );
 
+error:
     g_strfreev(keysize_strs);
+    g_strfreev(profile_names);
 
-    return 0;
+    return ret;
 }
 
 static int change_process_owner(const char *user)
@@ -1164,15 +1305,15 @@ error:
     return ret;
 }
 
-static int handle_create_config_files(const char *optarg)
+static int handle_create_config_files(const char *opt_arg)
 {
     g_auto(GStrv) tokens = NULL;
     gboolean overwrite = FALSE;
     gboolean root_flag = FALSE;
     gboolean skip_if_exist = FALSE;
 
-    if (optarg) {
-        tokens = g_strsplit_set(optarg, ", ", -1);
+    if (opt_arg) {
+        tokens = g_strsplit_set(opt_arg, ", ", -1);
         overwrite = g_strv_contains((const gchar **)tokens, "overwrite");
         root_flag = g_strv_contains((const gchar **)tokens, "root");
         skip_if_exist = g_strv_contains((const gchar **)tokens, "skip-if-exist");
@@ -1185,6 +1326,22 @@ static int handle_create_config_files(const char *optarg)
     return create_config_files(overwrite, root_flag, skip_if_exist);
 }
 
+static int read_config_file(const gchar *config_file,
+                            const struct passwd *user,
+                            gchar ***config_file_lines)
+{
+    if (access(config_file, R_OK) != 0) {
+        logerr(gl_LOGFILE, "User %s cannot read config file %s.\n",
+               user ? user->pw_name : "<unknown>", config_file);
+        return -1;
+    }
+
+    if (read_file_lines(config_file, config_file_lines))
+        return -1;
+
+    return 0;
+}
+
 int main(int argc, char *argv[])
 {
     int opt, option_index = 0;
@@ -1227,10 +1384,17 @@ int main(int argc, char *argv[])
         {"version", no_argument, NULL, '1'},
         {"print-capabilities", no_argument, NULL, 'y'},
         {"reconfigure", no_argument, NULL, 'R'},
+        {"profile", required_argument, NULL, 'I'},
+        {"profile-name", required_argument, NULL, 'J'},
+        {"profile-file", required_argument, NULL, 'g'},
+        {"profile-file-fd", required_argument, NULL, 'G'},
+        {"profile-remove-disabled", required_argument, NULL, 'j'},
+        {"print-profiles", no_argument, NULL, 'M'},
         {"help", no_argument, NULL, 'h'},
         {NULL, 0, NULL, 0}
     };
     unsigned long flags = 0;
+    g_auto(GStrv) config_file_lines = NULL;
     g_autofree gchar *swtpm_prg = NULL;
     g_autofree gchar *tpm_state_path = NULL;
     struct swtpm_backend_ops *backend_ops = &swtpm_backend_dir;
@@ -1243,17 +1407,23 @@ int main(int argc, char *argv[])
     g_autofree gchar *vmid = NULL;
     g_autofree gchar *pcr_banks = NULL;
     gboolean printcapabilities = FALSE;
+    gboolean printprofiles = FALSE;
     g_autofree gchar *keyfile = NULL;
     long int keyfile_fd = -1;
     g_autofree gchar *pwdfile = NULL;
     long int pwdfile_fd = -1;
     g_autofree gchar *cipher = g_strdup("aes-128-cbc");
-    g_autofree gchar *rsa_keysize_str = g_strdup_printf("%d", DEFAULT_RSA_KEYSIZE);
+    g_autofree gchar *rsa_keysize_str = NULL;
     unsigned int rsa_keysize;
     g_autofree gchar *swtpm_keyopt = NULL;
     g_autofree gchar *runas = NULL;
     g_autofree gchar *certsdir = NULL;
     g_autofree gchar *user_certsdir = NULL;
+    g_autofree gchar *json_profile = NULL;
+    g_autofree gchar *json_profile_name = NULL;
+    g_autofree gchar *json_profile_file = NULL;
+    g_autofree gchar *profile_remove_disabled_param = NULL;
+    int json_profile_fd = -1;
     gchar *tmp;
     gchar **swtpm_prg_l = NULL;
     gchar **tmp_l = NULL;
@@ -1264,7 +1434,7 @@ int main(int argc, char *argv[])
     char *endptr;
     gboolean swtpm_has_tpm12 = FALSE;
     gboolean swtpm_has_tpm2 = FALSE;
-    int fds_to_pass[1] = { -1 };
+    int fds_to_pass[2] = { -1, -1 };
     unsigned n_fds_to_pass = 0;
     char tmpbuffer[200];
     time_t now;
@@ -1414,6 +1584,7 @@ int main(int argc, char *argv[])
         case 'A': /* --rsa-keysize */
             g_free(rsa_keysize_str);
             rsa_keysize_str = strdup(optarg);
+            flags |= SETUP_RSA_KEYSIZE_BY_USER_F;
             break;
         case '3': /* --write-ek-cert-files */
             g_free(user_certsdir);
@@ -1438,6 +1609,39 @@ int main(int argc, char *argv[])
         case 'R': /* --reconfigure */
             flags |= SETUP_RECONFIGURE_F;
             break;
+        case 'I': /* --profile */
+            g_free(json_profile);
+            json_profile = g_strdup(optarg);
+            break;
+        case 'J': /* --profile-name */
+            g_free(json_profile_name);
+            json_profile_name = g_strdup(optarg);
+            break;
+        case 'g': /* --profile-file */
+            g_free(json_profile_file);
+            json_profile_file = g_strdup(optarg);
+            break;
+        case 'G': /* --profile-file-fd */
+            json_profile_fd = strtoull(optarg, &endptr, 10);
+            if (*endptr != '\0' || json_profile_fd >= INT_MAX) {
+                fprintf(stderr, "Invalid file descriptor '%s'\n", optarg);
+                goto error;
+            }
+            break;
+        case 'j': /* --profile-remove-disabled */
+            if (strcmp(optarg, "fips-host") != 0 &&
+                strcmp(optarg, "check") != 0) {
+                fprintf(stderr,
+                        "Unsupported parameter for --profile-remove-disabled: %s\n",
+                        optarg);
+                goto error;
+            }
+            g_free(profile_remove_disabled_param);
+            profile_remove_disabled_param = g_strdup(optarg);
+            break;
+        case 'M': /* --print-profiles */
+            printprofiles = TRUE;
+            break;
         case '?':
         case 'h': /* --help */
             usage(argv[0], config_file);
@@ -1465,13 +1669,12 @@ int main(int argc, char *argv[])
     }
     g_free(tmp);
 
-
-    ret = get_supported_tpm_versions(swtpm_prg_l, &swtpm_has_tpm12, &swtpm_has_tpm2);
+    ret = get_supported_tpm_versions((const char **)swtpm_prg_l, &swtpm_has_tpm12, &swtpm_has_tpm2);
     if (ret != 0)
         goto error;
 
     if (printcapabilities) {
-        ret = print_capabilities(swtpm_prg_l, swtpm_has_tpm12, swtpm_has_tpm2);
+        ret = print_capabilities((const char **)swtpm_prg_l, swtpm_has_tpm12, swtpm_has_tpm2);
         goto out;
     }
 
@@ -1489,6 +1692,22 @@ int main(int argc, char *argv[])
             goto error;
     }
 
+    curr_user = getpwuid(getuid());
+
+    if (printprofiles) {
+        ret = 0;
+        if (read_config_file(config_file, curr_user, &config_file_lines) < 0)
+            goto error;
+
+        if (flags & SETUP_TPM2_F) {
+            if (profile_printall((const char **)swtpm_prg_l, config_file_lines))
+                ret = 1;
+        } else {
+            printf("{}\n");
+        }
+        goto out;
+    }
+
     if (!got_ownerpass)
         ownerpass = g_strdup(DEFAULT_OWNER_PASSWORD);
     if (!got_srkpass)
@@ -1509,8 +1728,6 @@ int main(int argc, char *argv[])
         fclose(tmpfile);
     }
 
-    curr_user = getpwuid(getuid());
-
     // Check tpm_state_path directory and access rights
     if (tpm_state_path == NULL) {
         logerr(gl_LOGFILE, "--tpm-state must be provided\n");
@@ -1562,7 +1779,7 @@ int main(int argc, char *argv[])
     }
 
     if (!(flags & SETUP_RECONFIGURE_F)) {
-        ret = check_state_overwrite(swtpm_prg_l, flags, tpm_state_path);
+        ret = check_state_overwrite((const char **)swtpm_prg_l, flags, tpm_state_path);
         if (ret == 1) {
             goto error;
         } else if (ret == 2) {
@@ -1575,11 +1792,9 @@ int main(int argc, char *argv[])
             goto error;
     }
 
-    if (access(config_file, R_OK) != 0) {
-        logerr(gl_LOGFILE, "User %s cannot read config file %s.\n",
-               curr_user ? curr_user->pw_name : "<unknown>", config_file);
+    if (!config_file_lines &&
+        read_config_file(config_file, curr_user, &config_file_lines) < 0)
         goto error;
-    }
 
     /* check pcr_banks; read from config file if not given */
     tmp_l = g_strsplit(pcr_banks ? pcr_banks : "", ",", -1);
@@ -1590,7 +1805,62 @@ int main(int argc, char *argv[])
     g_strfreev(tmp_l);
     if (n == 0) {
         g_free(pcr_banks);
-        pcr_banks = get_default_pcr_banks(config_file);
+        pcr_banks = get_default_pcr_banks(config_file_lines);
+    }
+
+    if ((json_profile != NULL) +
+        (json_profile_name != NULL) +
+        (json_profile_file != NULL) +
+        (json_profile_fd > 0) > 1) {
+        logerr(gl_LOGFILE, "Only one of --profile, --profile-name, --profile-file, and --profile-file-fd may be given.\n");
+        goto error;
+    }
+
+    if (json_profile_name) {
+        if (profile_name_check(json_profile_name) < 0)
+            goto error;
+        /*
+         * Load profile from distro and local locations; sets json_profile_file
+         * to filename or json_profile with the JSON.
+         */
+        if (profile_get_by_name(config_file_lines,
+                                json_profile_name,
+                                &json_profile_file,
+                                &json_profile) < 0) {
+            logerr(gl_LOGFILE, "Could not find or access profile '%s'.\n",
+                   json_profile_name);
+            goto error;
+        }
+    }
+
+    if (json_profile_file) {
+        json_profile_fd = open(json_profile_file, O_RDONLY);
+        if (json_profile_fd < 0) {
+            logerr(gl_LOGFILE, "Could not open profile file '%s': %s\n",
+                   json_profile_file, strerror(errno));
+            goto error;
+        }
+    }
+
+    /* read default profile from swtpm_setup.conf */
+    if ((flags & SETUP_TPM2_F) != 0 &&
+        json_profile == NULL && json_profile_fd < 0) {
+        json_profile_fd = get_default_profile_fd(config_file_lines);
+        if (json_profile_fd == -2)
+            goto error;
+        if (json_profile_fd < 0)
+            json_profile = get_default_profile(config_file_lines);
+    }
+
+    if (json_profile_fd >= 0)
+        fds_to_pass[n_fds_to_pass++] = json_profile_fd;
+
+    if ((flags & SETUP_TPM2_F) != 0 && json_profile) {
+        if (validate_json_profile((const char **)swtpm_prg_l, json_profile) != 0)
+            goto error;
+    } else if (json_profile) {
+        logerr(gl_LOGFILE, "There's no --profile support for TPM 1.2\n");
+        goto error;
     }
 
     if (cipher != NULL) {
@@ -1631,11 +1901,14 @@ int main(int argc, char *argv[])
         logit(gl_LOGFILE, "  The TPM's state will be encrypted using a key derived from a passphrase (fd).\n");
     }
 
+    if ((flags & SETUP_RSA_KEYSIZE_BY_USER_F) == 0)
+        rsa_keysize_str = get_default_rsa_keysize(config_file_lines);
+
     if (strcmp(rsa_keysize_str, "max") == 0) {
         unsigned int *keysizes = NULL;
         size_t n_keysizes;
 
-        ret = get_rsa_keysizes(flags, swtpm_prg_l, &keysizes, &n_keysizes);
+        ret = get_rsa_keysizes(flags, (const char **)swtpm_prg_l, &keysizes, &n_keysizes);
         if (ret)
             goto error;
         g_free(rsa_keysize_str);
@@ -1652,7 +1925,7 @@ int main(int argc, char *argv[])
         size_t n_keysizes;
         gboolean found = FALSE;
 
-        ret = get_rsa_keysizes(flags, swtpm_prg_l, &keysizes, &n_keysizes);
+        ret = get_rsa_keysizes(flags, (const char **)swtpm_prg_l, &keysizes, &n_keysizes);
         if (ret)
             goto error;
 
@@ -1704,7 +1977,8 @@ int main(int argc, char *argv[])
     } else {
         ret = init_tpm2(flags, swtpm_prg_l, config_file, tpm_state_path, vmid, pcr_banks,
                        swtpm_keyopt, fds_to_pass, n_fds_to_pass, rsa_keysize, certsdir,
-                       user_certsdir);
+                       user_certsdir, json_profile, json_profile_fd,
+                       profile_remove_disabled_param);
     }
 
     if (ret == 0) {
